                LIST    P=16F84
                
                
INDIR   EQU     00H     	; Indirect pointer
RTCC    EQU     01H     	; Real time clock / counter
PCL     EQU     02H     	; Program counter low byte
STATUS  EQU     03H     	; Status register
C       EQU     0       	; Carry/borrow bit
DC      EQU     1       	; Digit carry
Z       EQU     2       	; Zero flag
PD      EQU     3       	; Power down
T0      EQU     4       	; Time-out bit
RP0     EQU     5       	; Register page select bits
RP1     EQU     6       	; Not used in 16C84
RP2     EQU     7       	; Not used in 16C84
FSR     EQU     04H     	; Indirect pointer
PORTA   EQU     05H     	; PORT A
PORTB   EQU     06H     	; PORT B
EEDATA  EQU     08H     	; Eeprom data
EEADR   EQU     09H     	; Eeprom address
PCLATH  EQU     0AH     	; Program counter high byte
INTCON  EQU     0BH     	; Interrupt control
TRISA   EQU     085H    	; Data direction port A
TRISB   EQU     086H    	; Data direction port B
EECON1  EQU     088H    	; Eeprom control
EECON2  EQU     089H    	; Eeprom control
W       EQU     0       	; W reg. as destination
F       EQU     1       	; F file as destination

		
		__CONFIG	3ff1H
		
;********************************************************************************
; CONDIZIONI DI COMPILAZIONE

; Impostare a 1 se si desidera compilare per questo decoder
HUMAX5400       equ     0

; Impostare a 1 se si desidera compilare per wafer smd
SMD		equ	1

;********************************************************************************
; Stringhe di sostituzione per test&skip sui bit di zero o carry

#DEFINE skipz   btfss   STATUS,Z
#DEFINE skipnz  btfsc   STATUS,Z
#DEFINE skipc   btfss   STATUS,C
#DEFINE skipnc  btfsc   STATUS,C
#DEFINE carrybit  	STATUS, 000H

;---------------------------------------------
; Definizioni delle variabili
;
rb4     	equ     04H     	; pin 10
rb5     	equ     05H      	; pin 11     usato dal soft originario.
rb7     	equ     07H     	; pin 13
iocam   	equ     rb7     	; i/o CAM
sda     	equ     rb4     	; serial data eeprom esterna
scl     	equ     rb5		; serial clock eeprom esterna
BIT_DELAY	equ	23

;---------------------------------------------
; Definizioni delle variabili
;
; howbyte -> 		bit 7 = 0-calc key 	1-calc signa
;			bit 6 = 0-prov 00 	1-prov 10
;			bit 5 = 0-		1-agg. pmk
;			bit 4 = 0-		1-agg. key
;			bit 3 = 0-command scan	1-command exec
;			bit 2 = 0-ricerca chid	1-nessuna ricerca
;			bit 1 =	0-		1-gestione PIN
;                       bit 0 = 0-prima key 	1-seconda key (classe 5 [12/13H])
;
	
	CBLOCK  0CH
	  rxtxbuf		; buffer di ricezione trasmissione seriale
	  rxtxcnt           	; contatore dei bit usato nel protocollo seriale
	  howbyte         	;
	  temp2           	; registro temporaneo
	  temp1           	; registro temporaneo
	  index			; indice del provider corrente
	  keynum          	; numero della chiave	  
 	  adr_h			; appoggio indirizzi alti eeprom esterna
	  adr_l			; appoggio indirizzi bassi eeprom esterna
	  bytecnt           	; contatore generico

	  prov            	; provider (00/01)
	  bf2_cla         	; contiene la classe del comando
	  bf3             
	  bf4             
	  bf5             
	  bf6             
	  bf7
	  bf8_lun         	; contiene il numero di byte da ricevere ancora
	  chid_h          	; channel id alto
	  chid_l          	; channel id basso
	ENDC			; 00Ch - 01Fh
	
	CBLOCK  20H
	  sbuffer		; 020H-045H (38 BYTE) signcalc buffer
	 ENDC
	  
        CBLOCK 046H
	  tmpkeybuf       	; 046H-04DH buffer temporaneo
	ENDC
	
	CBLOCK 04EH
	  date_h            	; data parte alta
	  date_l            	; data parte bassa
	ENDC
	  
chksum		equ	prov		; ex. chid_h - contiene lo xor linkage della trasmissione
taboff          equ	chid_l		; spiazzamento nella look-up table

pin_h		equ	sbuffer
pin_l		equ	sbuffer + 001H
pin_pclath	equ	sbuffer + 021H
pin_counterw	equ	sbuffer + 022H
pin_fsr		equ	sbuffer + 023H
pin_check	equ	sbuffer + 024H
pin_counter	equ	sbuffer + 025H

i		EQU	rxtxbuf			; solo per similitudine src in c++
si		EQU	rxtxcnt			; solo per similitudine src in c++

;--------------------------------------------------------------------------------
;--------------------------------------------------------------------------------

		ORG     0000H

init		GOTO    main		; vai al punto di entrata


;*************************************************************************************
;GESTIONE COMANDI CLASSE 2
;*************************************************************************************

is_00		movlw   015H		; classe 2, IS=00 -> serial ascii
     		call	read_max_buffer
     		movlw	014H
		goto	send_buffer_0
		
is_01		movlw	02AH		; classe 2, IS=01 -> serial hex
		call	read_max_buffer

		call	pagee_02
		bcf	carrybit	
		rlf	index, W
		addwf	index, W	
		addlw	0C0H	
		movwf	EEADR		; ex. taboff
		movlw	sbuffer + 00CH
		call	read_3_bytes

		movlw	010H
		goto	send_buffer_0
	
is_02		movlw   03BH		; classe 2, IS=02 -> country code
		call	read_max_buffer
		
		movlw	0F0H		; pos coco
		addwf	index, W
		movwf	EEADR
		call	read_ees
		movwf	sbuffer + 006H	; pos coco
		
		movlw	010H
		goto	send_buffer_0

is_03		movlw	04CH		; classe 2, IS=03 -> provider id
		btfsc	bf7,00H         ; skip if bit = 0
		movlw	065H
		call	read_max_buffer
	
		call	pagee_07	
		addlw	00AH
		btfsc	bf7,00H         ; skip if bit = 0
		addlw	003H
	
		movwf	EEADR		; ex. taboff
		movlw	sbuffer + 001H
		call	read_3_bytes

		movlw	018H
		goto	send_buffer_0

is_08		movlw	003H		; classe 2, IS=08 (Humax)
		movwf	bf6		; trasmette tutti zeri!  
		movlw	020H
		movwf	bf8_lun
		goto	is_z_talk

is_09		movwf	EEADR		; ex. taboff; classe 2, IS=09 -> send CAM key
		movlw	055H
		movwf	bf3
		clrf	bf8_lun
		call	write_ee_8      ; scrive la camkey in eeprom esterna (ripetuto ogni volta)
		goto	send_buffer_1
		

is_0A		clrf    bf8_lun		; classe 2, IS=0A -> PIN

		swapf	pin_h, W
		andlw	00FH
		xorlw	009H
		skipnz
		goto	init_prg

		btfss	howbyte, 001H
		goto	ok_pin
		
		movf	bf6,F
		skipz
		goto	pin_error

		movf	pin_fsr, W
		movwf	FSR	
		
		movf	pin_l, W
		addwf	pin_h, W		
		movwf	INDIR
		xorwf	pin_check, F
		incf	pin_fsr, F
		
		decfsz	pin_counter, F
		goto	pin_error
				
		movf	pin_check, F
		skipz
		goto	ok_pin
		
		movf	pin_pclath, W
		movwf	PCLATH
		
		movf	sbuffer + 0AH, W	; indirizzo di inizio scrittura
		call    write_ee_0      	; init scrittura
		movf	pin_counterw, W
		movwf	bytecnt
		movlw	sbuffer + 0BH
		call    write_ee_w      ; copia in eeprom da sbuffer+0BH in poi
		
		goto	ok_pin		

init_prg	movf	pin_l, W
		addwf	pin_h, W
		movwf	pin_check
		
		btfsc	pin_h, 003H
		goto	pin_sel_prov		
		
		movf	pin_h, W
		andlw	07H
		movwf	index
		
		swapf	pin_l, W
		andlw	00FH		
		movwf	pin_pclath
		
		movf	pin_l, W
		andlw	00FH
		movwf	pin_counterw
		addlw	002H
		movwf	pin_counter
		
		movlw	sbuffer + 00AH
		movwf	pin_fsr
		
		bsf	howbyte, 001H		;goto	pin_error		

pin_error	movlw   051H
		goto	end_pin
	
pin_sel_prov	movf	pin_l, W
		movwf	index
		bcf	howbyte, 002H
		btfss	index, 003H
		bsf	howbyte, 002H
		
		call	aggiorna_F9

ok_pin		bcf	howbyte, 001H
		movlw	050H
		movf	bf6,F
		skipz
		movlw	05EH
end_pin		movwf	bf3	
		goto	send_buffer_1
		
		
is_0B		movlw	07EH		; classe 2, IS=0B -> ????????
		call	read_max_buffer
		movlw	02BH
		goto	send_buffer_0

is_11		movlw	058H		; classe 2, IS=11 
		movwf	bf3		; trasmette tutti zeri!
					; goto	is_0E

is_0E		movlw	040H		; classe 2, IS=0E -> card file
		movwf	bf8_lun		; trasmette tutti zeri!
is_z_talk	call	talk

is_zero		movlw	000H            
		call	txcam
		decfsz	bf8_lun,F       ; skip if Zero
		goto	is_zero		
		call	talk_chk	
		goto	dloop

read_3_bytes	movwf	FSR
		movlw	003H
		movwf	bytecnt	
		goto	read_in_sbufc

cl2_gst		movf	bf5, W
		movwf	EEADR
		call	read_i_byte
		clrf	PCLATH
		movwf	PCL		

read_i_byte	bsf	STATUS, RP0
		bsf	EECON1, 000H
		bcf	STATUS, RP0
		movf	EEDATA, W
		return


read_max_buffer	movwf	EEADR
		movlw	025H		
		goto	read_in_sbuf

send_buffer_0	movwf	bf8_lun
send_buffer_1	call	talk
send_buffer_2	call	send_sbuffer
		call	talk_chk
jmpdloop	goto	dloop


;---------------------------------------------
; RITARDO
; tempo di ritardo (4/6MHz)*((256-W)*4)=(256-W)*2.66us

delay		addlw	001H
		btfss   STATUS,Z        ; skip if bit = 1 (esce se zero)
		goto    delay           ; ripete -> e' un loop di ritardo
		return

;---------------------------------------------
; TRASMISSIONE
					; trasmissione alla CAM
txcam          
	MOVWF   rxtxbuf         
	XORWF   chksum,F
	MOVLW   008H
	MOVWF   rxtxcnt
	MOVLW   0E0H
	CALL    delay

txcam1
	BTFSS   PORTB,iocam     ; skip if bit = 1
	GOTO    txcam1          ; attende linea libera
	BSF     STATUS,RP0      ; configura iocam come uscita
	BCF     PORTB,iocam
	BCF     STATUS,RP0 
	BCF     PORTB,iocam 
	MOVLW   0DBH
	CALL    delay

txcam2
	RRF     rxtxbuf,F       
	BTFSC   STATUS,C        ; skip if bit = 0
	BSF     PORTB,iocam 
	BTFSS   STATUS,C        ; skip if bit = 1
	BCF     PORTB,iocam 
	MOVLW   0DCH
	CALL    delay
	DECFSZ  rxtxcnt,F       ; skip if Zero
	GOTO    txcam2
	NOP
	NOP
	BSF     PORTB,iocam     ; iocam come input
	BSF     STATUS,RP0 
	BSF     PORTB,iocam 
	BCF     STATUS,RP0 
	MOVLW   0D0H
	;CALL    delay
	;RETURN
	goto 	delay

;-------------------------------------------------------------------
; RICEZIONE DA CAM


rxcam		MOVLW   008H            
		MOVWF   rxtxcnt

rxcam1		BTFSS   PORTB,iocam     ; skip if bit = 1
		GOTO    rxcam1

rxcam2		BTFSC   PORTB,iocam     ; skip if bit = 0
		GOTO    rxcam2
		MOVLW   0CAH
		CALL    delay
		NOP

rxcam3		BCF     STATUS,C      
		BTFSC   PORTB,iocam     ; skip if bit = 0
		BSF     STATUS,C 
		RRF     rxtxbuf,F
		MOVLW   0DCH
		CALL    delay
		NOP
		DECFSZ  rxtxcnt,F       ; skip if Zero
		GOTO    rxcam3

		IF HUMAX5400
		 MOVLW   0FFH		; accorciato stop bit
		ELSE
		 MOVLW   0E8H            ; stop bit normale
		ENDIF

		CALL    delay
		MOVF    rxtxbuf,W
		XORWF   chksum, F
		RETURN

;----------------------------------------------------------------

					; legge 8 byte dalla cam e li salva nell'indirizzo
					; contenuto in W all'entrata della routine
rx_save
	MOVWF   FSR             
	MOVLW   008H

rx_save1
	MOVWF   EEADR
           
rx_save_001
	CALL    rxcam          
	MOVWF   INDIR
	INCF    FSR,F
	DECFSZ  EEADR,F		; skip if Zero
	GOTO    rx_save_001
	RETURN

;-----------------------------------------------------------------

aggiorna_F9	call	read_index
		xorwf   index, W	; confronta con index	
		skipnz			; se uguali ritorna
		return

write_ee_i	movlw	0F9H
		call	write_ee_0	; salva index in eeprom esterna 
		movf	index,W
write_ee_i2	call	wr_ee_est
		goto	eb_ee_est
		
read_index	call	pagee_02
		movlw	0F9H
		;call	read_ee
		;return
		goto	read_ee

;-------------------------------------------------------------------
				; programma principale
main

		;bsf	howbyte, 006
		;movlw	000H
		;movwf	index
		;movlw	009H
		;movwf	keynum
		;call	find_addr
		;goto	prova1
		


		;MOVLW   0FFH            ;A tutti i pin come ingressi
		;TRIS    PORTA
		;TRIS    PORTB
		;MOVWF   PORTA           ; mette ffH anche nel buffer dei port
		;MOVWF   PORTB                
                ;MOVLW   07FH            ; carica 7FH in W e lo mette nel registro OPTION
		;OPTION
		;BSF     PORTB,iocam		
		
		bsf	STATUS,RP0	;Select	page 1 (80$ - FF$)
		bcf	RTCC,07		;Set PULL-UPs, rising edge INT,	CLKOUT,	increment on low to high, prescaler for	Wdog
		movlw	0ffh		;All Input
		movwf	PORTB		;Set portb 0FF$
		bcf	STATUS,RP0	;Select	page 0 (00$ - 7F$)


		call	pagee_02
                movlw	014H		; si prepara per la trasmissione dell'ATR
		movwf	bytecnt
		clrf	EEADR
main_001	call	read_ees
		call	txcam
		incf	EEADR, F		
		decfsz	bytecnt, F
		goto	main_001                
                
                call	read_index
		movwf	index
		clrf	howbyte

dloop		CLRF    bf3             ; pulisce i bf3 e bf4
		CLRF    bf4
		;CALL    listen		
		
		;-------------------------------------------------------------------
		; routine di ricezione protocollo standard dalla CAM
		
 
listen		MOVLW   03FH            ; LISTEN
		MOVWF   chksum
		CALL    rxcam           ; ricevi un byte da ... (dipende da devsel)   (1) [1]
		XORLW   001H            ; controlla il byte in W (quello ricevuto) sia 01H
		skipz
		GOTO    listen
		CALL    rxcam           ; ricevi un altro byte  (2) [2]
		MOVWF   bf2_cla         ; salvalo in bf2_cla
	
		XORLW   005H            ; e' stato ricevuto 005H ?
		skipnz
		GOTO    listen_ok
		XORLW   007H            ; e' stato ricevuto 002H ?
		skipnz
		GOTO    listen_ok
		XORLW   003H            ; e' stato ricevuto 001H ?
		skipnz
		GOTO    listen_ok
		GOTO    listen          ; errore,ricomincia !!!

listen_ok	movlw	bf5		; riceve 4 byte -> mem. bf5-bf6-bf7-bf8_lun
		movwf	FSR
		movlw	004H
		call	rx_save1
	
		MOVF    bf2_cla, W     	; testa la classe
		XORLW   002H           	; classe 2
		skipz
		goto	main_lis_ret
		
		MOVF    bf8_lun,F      	; testa se l'ultimo byte e' zero
		skipnz
		GOTO    listen_chk
	
		MOVF    bf5, W      	; testa se cam key (is=09)	
		XORLW   009H           	;
		skipz
		GOTO    listen_001
	
		incf	bf7, F

listen_003	MOVLW   008H           	; salva la cam key 
		SUBWF   bf8_lun, F
		MOVLW   tmpkeybuf      
		CALL    rx_save
		decfsz	bf7, F
		goto	listen_003
		
listen_001	clrf	bytecnt
		incf	bf8_lun, F
		goto	cestina2

listen_end	CALL    rxcam 		; riceve e cestina la rimanenza	
		btfss	bytecnt, 000H
		movwf	pin_h
		movwf	pin_l
		movlw	001H
		xorwf	bytecnt, F
	
cestina2	DECFSZ  bf8_lun, F      ; skip if Zero
		GOTO    listen_end

listen_chk	CALL    rxcam
		MOVF    chksum, F
		skipz
		GOTO    listen
		;RETURN
		;-------------------------------------------------------------------
		
		
main_lis_ret	btfss	bf2_cla, 000H	;se classe 1 o classe 5 vai a command_scan
		goto	cl2_gst
		
;*************************************************************************************
;GESTIONE COMANDI 
;*************************************************************************************
		
command_scan_0	
		bcf	howbyte, 004H
		movlw	sbuffer
		movwf	FSR		
		movlw	006H
		call	rx_save1	; riceve i primi sei byte
		
		movlw	00BH
		subwf	bf8_lun, F
		movf	bf8_lun, W		
		movwf	taboff
		movwf	bf4		
				
fltr_ciclo	movlw	001H
		call	rx_save1	; riceve nano
		movwf	temp1
		
		movlw	001H
		call	rx_save1	; riceve lunghezza
		andlw	03FH
		movwf	temp2
		call	rx_save1	; riceve blocco dati nano
		
		movlw	01CH
		subwf	bf4, W
		skipc
		goto	fltr_ok_nano

		;***** superara la lunghezza		
		
		movf	temp1, W
		andlw	03FH
		skipnz
		goto	fltr_ok_nano
		
		xorlw	038H
		skipnz
		goto	fltr_ok_nano	
		
		movf	temp2, W
		addlw	002h		
		subwf	FSR, F
		subwf	bf8_lun, F
		subwf	sbuffer+5, F
		
		bsf	howbyte, 004H	; setta per superare il controllo signa

fltr_ok_nano	movf	temp2, W
		subwf	taboff, F
		decf	taboff, F
		decfsz	taboff, F
		goto	fltr_ciclo
		
		movlw	006H
		call	rx_save1
		movf	chksum, F
		skipz
		goto	dloop		
		
		;movf	bf4, W
		;addlw	00BH
		;movwf	bf8_lun		
		;call	send_sbuffer
		;goto	dloop
		
				
newcl1p		;movlw	00BH
		;subwf	bf8_lun, F

	  	;*********************
	  	; INIZIO SCANSIONE COMANDO
	  	; si=6 - prov=Lunghezza totale nanos	  	

		bcf	howbyte, 003h

		movlw	006h
		movwf	si
				
		bcf	howbyte, 006H	; cancella flag prov. 10

command_scan00	bcf	howbyte, 005H	; cancella flag agg pmk
		clrf	taboff
			  	
command_scan	call	ee_int_rib	; legge il nanocomando
	  	movwf	EEDATA		; in EEDATA
	  	call	ee_int_rib	; legge la lunghezza
	  	movwf	i		; in i
  		
  		movlw	03fH
		andwf	i, F
	
		movlw	002H
		addwf	i, W	
	  	subwf	bf8_lun, F

		movlw	03FH
		andwf	EEDATA, W	; nano agg. data
		skipnz
		goto	agg_date
		
		xorlw	010H		; nano agg. key		(10)
		skipnz
		goto	agg_key
		
		xorlw	038H		; nano agg. pmk		(28)
		skipnz
		goto	agg_pmk

		xorlw	010H		; nano richiesta key	(78)
		skipnz
		goto	cs_classe5

nano_unknow	call	ee_int_rib	; aggiorn. Data
		decfsz	i, F
		goto	nano_unknow
		goto	next_nano		
		
agg_date	call	ee_int_rib	; aggiorn. Data
		movwf	date_h
		call	ee_int_rib
		movwf	date_l
		
		movf	si, w
		movwf	prov
		movf	bf8_lun, W
		movwf	temp1		; keynum		
		goto	next_nano		
		
agg_pmk		bsf	howbyte, 5	; flag agg. pmk
		call	ee_int_rib	;
		btfsc	W, 000H
		bsf	howbyte, 006H	; flag prov. 10		
		
		call    pagee_07        ; pagina 7 ( swapf index,W )
		addlw   008H		; calcola indirizzo penultimo byte HMK
		call	read_ee         ; indirizzo 0700|(indexswapped+08)
		movwf   date_h
		incf	EEADR, F
		call	read_ees
		movwf   date_l

		movf	si, W
		movwf	temp1		; temp1 in questo caso = offset MK - 2
		
		decf	i, F		; toglie 1 byte
		
		goto	agg_key2

agg_key		bsf	howbyte, 004H	; flag agg. key

		btfsc	taboff, 000H
		goto	agg_key_c

		btfsc	sbuffer, 003H
		bsf	howbyte, 006H	; flag prov. 10		

agg_key2	btfss	howbyte, 003h
		call	keep_pkey_0
		
agg_key_c	call	ee_int_rib
		movwf	keynum		; numero key
		
		movf	FSR, W
		addlw	001H		; offset della key in sbuffer in W
		movwf	temp2
		
		btfss	howbyte, 003h
		goto	cmd_scan
		
		;***************************
		;ESECUZIONE COMANDO AGG KEY
		
		call	cmd_agg_keys
		goto	no_scan
		
		;***************************		

cmd_scan	call	seskeycrypt2
		call	cs_classe5_s1
		bsf	taboff, 000H

no_scan		btfss	howbyte, 5	;salta se settato il bit 5
		goto	agg_key_c1
		
		movf	i, W
		addwf	si, F
		goto 	next_nano		
		
agg_key_c1	movlw	008H
		addwf	si, F
		subwf	i, F
		decfsz	i, F
		goto	agg_key_c
		goto	next_nano		

		;************************* CLASSE 5

cs_classe5_s1	call	rot_right10
		call	rot_right10
		call	rot_right10		
		call	copyleft_10
		call	copyleft_10
		movlw	008H
		addwf	temp2, F
		return

cs_classe5	btfsc	sbuffer + 2, 004H
		bsf	howbyte, 006H	; flag prov. 10
		
		call	ee_int_rib	; numero key
		movwf	keynum
		call	ee_int_rib	; byte discriminante 12/13 (prima o seconda key)
		bcf	howbyte, 0	; cancella il bit 0 di howbyte
		btfsc	W, 000H
		bsf	howbyte, 0	; setta il bit 0 di howbyte (richiesta seconda key)
		
		incf	FSR, W
		movwf	temp2		;offset della key in sbuffer

		call    find_addr       ; trova l'indirizzo della pkey		
		
		call	seskeycrypt	; decrypta la prima key
		call	cs_classe5_s1		
		call	seskeycrypt2	; decrypta la seconda key
		call	cs_classe5_s1
		
		movlw	010H
		addwf	si, F
		subwf	temp2, W
		movwf	keynum		; offset delle key

next_nano	movf	bf8_lun, F
		skipz
		goto	command_scan
		
		btfsc	howbyte, 003h
		goto	cl1_end

		
		;*********************
	  	; INIZIO CALCOLO SIGNA
	  	; si = 0
	  	
	  	movf	sbuffer+5, W
		addlw	001H
		movwf	bf8_lun			;lunghezza totale -5 (signa)	

		movf	sbuffer, W
	  	movwf	chid_h	  	
	  	movf	sbuffer+1, W
	  	movwf	chid_l

		clrf	si
		bsf	howbyte, 7		;settare questo bit per calc. signature
		movlw	sbuffer		
		movwf	temp2		

cliclo_sign	movlw	028H
		call	newcrypt
		call	rot_right10
		
		movlw	008H
		addwf	si, F
		subwf	bf8_lun, F
		skipnz
		goto	fine_sign
		
		movlw	0F8H		; FF - 07
		andwf	bf8_lun, W
		skipnz
		goto	ultimo_blocco

		movlw	008H
		movwf	bf3		
		call	bk_xor_copy
		goto	cliclo_sign

ultimo_blocco	movf	bf8_lun, W
		movwf	bf3		
		call	bk_xor_copy		
				
		movlw	008H
		movwf	bf3
		movf	bf8_lun, W
		subwf	bf3, F
		addwf	si, F
		
		movlw	sbuffer
		addwf	bf8_lun, W
		movwf	FSR
		
		movlw	061H
completa_bk	xorwf	INDIR, F
		incf	FSR, F
		addlw	001H
		decfsz	bf3, F
		goto	completa_bk

		movlw	068H
		call	newcrypt

fine_sign	clrf	i
		movlw	005H
		movwf	bf3		
		call	bk_xor_copy	

		movf	i, F
		skipnz
		goto	command_ok


		;****************************
		;SIGNA ERRATA		

signa_err	movlw	028H
		btfss   bf2_cla, 002H		
		goto	cl1_end2
		
		btfss	howbyte, 002
		call    sel_prov        ; seleziona il provider		
		
		btfss	howbyte, 004h	; comando lungo, signa sempre errata
		goto	cl5_nok
				
		;****************************
		;SIGNA OK

command_ok	btfsc   bf2_cla,02H     ; skip if bit 2 = 0 (se non  cl. 5)
		goto	cl5_ok

aggiorna_key	btfsc	howbyte, 5
		goto	aggiorna_pmk		

aggiorna_pk
		bsf	howbyte, 003H
		movf	prov, W
		movwf	si
		movf	temp1, W
		movwf	bf8_lun
		goto	command_scan00

cmd_agg_keys	movwf	FSR
		call    find_addr       ; trova l'indirizzo della key attuale da aggiornare
		movf    adr_h,W         ; indirizzo alto caricato da adr_h
		movwf   PCLATH
		movf    adr_l,W         ; indirizzo basso da adr_l
		call    write_ee_0      ; inizializza la scrittura
		call	load_8_cnt
		call    write_ee_z 	; scrive 8 byte
		goto	bigdelay

aggiorna_pmk	
		call    pagee_07        ; pagina 7 ( swapf index,W )
		addlw   080H            ; indirizzo 0700|(indexswapped+80)
		btfsc	howbyte, 006H	; flag prov. 10		
		addlw	008H
		call    write_ee_0      ; init scrittura		
		call	load_8_cnt
		movf	temp1, W	; ( MOVLW   keyinbuf1 )
		addlw	sbuffer + 1
		call    write_ee_w      ; copia in eeprom
		call    bigdelay        ; ritardo 
		
		call    pagee_07        ; pagina 7 ( swapf index,W )
		addlw   00AH            ; offeset + 00AH
		btfsc	howbyte, 006H	; flag prov. 10		
		addlw	003H
		call    write_ee_0      ; init scrittura nuovo provider id
		movlw   003H
		movwf   bytecnt
		movf	temp1, W	; ( MOVLW   key_2 )
		addlw	sbuffer + 9
		call    write_ee_w      ; copia in eeprom da key_2 in poi
		call    bigdelay        ; ritardo 

;-------------------------------------------------------------------

cl1_end		movlw	020H
cl1_end2

risp_standard	movwf	EEADR
		movlw	bf2_cla
		movwf	FSR
		call	load_8_cnt

ciclo_read_i	call	read_i_byte
		movwf	INDIR
		incf	EEADR, F
		incf	FSR, F
		decfsz	bytecnt, F
		goto	ciclo_read_i
		clrf	bf8_lun
		goto	send_buffer_1

ee_int_rib	movlw	sbuffer
		addwf	si, W
		movwf	FSR
		movf	INDIR, W
		incf	si, F
		RETURN
		
bk_xor_copy	movlw	sbuffer - 1	; xor tra il primo blocco di 8 byte
		addwf	si, W		; e quello puntato da si
		addwf	bf3, W		; bf3 contiene il numero di byte da xorare
		movwf	FSR
		movf	INDIR, W
		movwf	EEADR		; temp1
		movf	si, W
		subwf	FSR, F
		movf	EEADR, W	;temp1, W
		xorwf	INDIR, F
		movf	INDIR, W
		
		iorwf	i, F
		;addwf	i, F
		
		decfsz	bf3, F
		goto	bk_xor_copy
		return

		
;-------------------------------------------------------------------
cl5_nok		movlw	030H		
		goto	risp_standard
		
cl5_ok
		movf	keynum, W
		movwf	temp2
		call	cardcrypt
		movlw	008H
		addwf	temp2, F		
		call	cardcrypt_001b
				
		movlw	0E0H	
		movwf	EEADR		; ex. taboff
		movlw	00DH
		call	read_in_sbuf
				
		btfss   howbyte,00H     ; skip if bit = 1
		goto	rispcont5

secondakey5	bsf	sbuffer + 004H, 000H
		bsf	sbuffer + 00AH, 000H

rispcont5	movf	chid_h, W
		movwf	sbuffer + 007H
		movf	chid_l, W
		movwf	sbuffer + 008H
				
		movlw	sbuffer - 1
		movwf	FSR		
		movlw	00EH
		movwf	EEADR
		call	talk_000

		movf	keynum, W
		movwf	FSR
		movlw	010H
		movwf	EEADR
		call	send_sbuffer2
		goto    dloop
		
;-----------------------------------------------------------------------------
; SCRIVE-LEGGE NELLA EEPROM ESTERNA
					; ecco come funziona la 24C16:
					; SLAVE ADDRESS:
					; bit7=1
					; bit6=s2
					; bit5=s1
					; bit4=s0
					; bit3=A10
					; bit2=A9
					; bit1=A8
					; bit0=read/write (1=read,0=write)

sa_ee_est       
	RLF     PCLATH,W        ; nel pclatch basso ci sono i tre bit dell'indirizzo alto 
					; ruota a sinistra cosi' nell'ultimo bit entra il carry che indica R/W
	ANDLW   00FH            ; tiene solo i bit piu' bassi
	IORLW   0A0H            ; 0A indirizza la eeprom
					; con s0=0 s1=0 s2=0 (nelle wafer gold s1 e' negato !!!!!!)
					; in W c'e' lo slave address

wr_ee_est
	MOVWF   EEDATA          ; salva W in EEDATA (registro di trasmissione)
	MOVLW   0CFH            ; rende uscite i rb4 (SDA) e rb5 (SCL)... (che vanno alti!!!)
	TRIS    PORTB
	CALL    wr_ee_1         ; trasmette 8 bit (4+4)
	CALL    wr_ee_1
	GOTO    br_ee_est

wr_ee_1
	CALL    wr_ee_2         ; esegue 4 volte wr_ee_2 
	CALL    wr_ee_2
	CALL    wr_ee_2
wr_ee_2
	RLF     EEDATA,F       
	GOTO    bw_ee_est1

rd_ee_est
	MOVLW   0DFH            ; rb4 (SDA) come ingresso
	TRIS    PORTB
	CALL    rd_ee_1         ; legge 8 bit (4+4) da rb4 (sda)
	CALL    rd_ee_1
	BSF     carrybit
	GOTO    bw_ee_est

rd_ee_1
	CALL    rd_ee_2           
	CALL    rd_ee_2
	CALL    rd_ee_2
rd_ee_2
	CALL    br_ee_est1          
	RLF     EEDATA,F
	RETURN

sb_ee_est
	BSF     PORTB,sda       ; setta rb4 (sda)
	MOVLW   0CFH            ; rb4 e rb5 uscite
	TRIS    PORTB
	BSF     PORTB,scl       ; alza il clock
	NOP
	BCF     PORTB,sda       ; abbassa SDA-> condizione di START
	GOTO    end_ee_2

eb_ee_est
	BCF     PORTB,sda       ; resetta rb4 (sda)
	MOVLW   0CFH            ; rb4 e rb5 uscite
	TRIS    PORTB
	NOP
	BSF     PORTB,scl       ; alza il clock
	NOP
	BSF     PORTB,sda       ; alza SDA -> condizione di STOP

end_ee_1
	NOP            	  	; ritorna 
end_ee_2
	BCF     PORTB,scl       ; porta basso scl 
	RETURN

bw_ee_est       
	MOVLW   0CFH            ; trasmette un bit (rb4 come uscita (sda))
	TRIS    PORTB

bw_ee_est1
	skipc				; trasmette un bit (rb4 come il carry
	BCF     PORTB,sda
	skipnc
	BSF     PORTB,sda
				; i dati cambiano con il clock basso!
	BSF     PORTB,scl       ; alza rb5 (scl) clock alto
	GOTO    end_ee_1        ; ritorna

br_ee_est
	MOVLW   0DFH            ; riceve un bit (rb4 come ingresso)
	TRIS    PORTB

br_ee_est1
	BSF     PORTB,scl       ; alza rb5 (scl) clock alto 
	NOP
	BSF     carrybit     	; setta il carry in accordo con lo stato di sda
	BTFSS   PORTB,sda       ; skip if bit = 1
	BCF     carrybit
	GOTO    end_ee_2

read_ee
	MOVWF   EEADR           ; salva W in EEADR

read_ees
	call	write_ee_01
	
	CALL    sb_ee_est       ; altro start 
	BSF     STATUS,C        ; leggo dalla eeprom
	CALL    sa_ee_est       ; slave address
	CALL    rd_ee_est       ; leggo un byte
	CALL    eb_ee_est       ; stop
	MOVF    EEDATA,W
	RETURN
                                       
write_ee_0			; inizializza la scrittura nella eeprom esterna
	MOVWF   EEADR           ; salva l'indirizzo 
write_ee_01	
	CALL    sb_ee_est       ; condizione di start
	BCF     carrybit    	; scrivo nella eeprom
	CALL    sa_ee_est       ; slave address        
	MOVF    EEADR,W
	GOTO    wr_ee_est       ; trasmetto l'indirizzo E RITORNA 

write_ee_8				; scrive 8 byte in eeprom esterna
	CALL    pagee_02		; setta indirizzo alto  
	MOVLW   0B0H            ; indirizzo basso (indirizzo camkey 02b0)
	CALL    write_ee_0      ; inizia la scrittura 
					; continua e scrive 8 byte

write_ee_x				; salva in eeprom 8 byte a partire da cambuf
	MOVLW   tmpkeybuf      
	MOVWF   FSR
	call	load_8_cnt
                                      
write_ee_y				; scrive bytecnt byte in eeprom estrna
	MOVF    INDIR,W         ; carica e salva in eeprom 
	CALL    wr_ee_est       
	INCF    FSR,F
	DECFSZ  bytecnt,F       ; skip if Zero
	GOTO    write_ee_y
	GOTO    eb_ee_est       ; end bit e ritorna 

write_ee_w 
	MOVWF   FSR            
write_ee_z
	GOTO    write_ee_y      

;-------------------------------------------------------------------
					; ruota tmpkeybuf a sinistra di 8 bit
					; il buffer e' sistemato cosi
					; byte 0,1...7
					; per ogni byte  i bit sono 7...0
copyleft_10	movlw	009H
		goto	copyleft_81
		
copyleft_8	movlw	007H
copyleft_81	movwf	EEDATA

copyleft	movlw	tmpkeybuf
		movwf	FSR
		movf	INDIR, W
		movwf	EEADR		; temp1		

copyleft1	incf	FSR, F
		movf	INDIR, W
		decf	FSR, F
		movwf	INDIR		
		incf	FSR, F
		decfsz	EEDATA,F
		goto	copyleft1
		movf	EEADR,W		; temp1, W
		movwf	INDIR
		return
		
;-------------------------------------------------------------------
					; ruota tmpkeybuf a destra di 8 bit
					; il buffer e' sistemato cosi
					; byte 0,1...7
					; per ogni byte  i bit sono 7...0
rot_right8	MOVLW   008H           
		GOTO    rot_right_in

rot_right10	MOVLW   00AH         

rot_right_in	MOVWF   EEDATA
		MOVLW   tmpkeybuf
		MOVWF   FSR
		BCF     carrybit

rot_right_c	RRF	INDIR, F
		INCF	FSR, F
		DECFSZ  EEDATA,F
		GOTO	rot_right_c
		MOVWF	FSR
		MOVLW	080H
		BTFSC	carrybit
		IORWF	INDIR, F
		RETURN
		
;-------------------------------------------------------------------
                                      
find_addr				; calcola l'indirizzo della pkey

	bcf	keynum, 000H
	
	movlw	011H
	movf	keynum, f
	skipnz
	movwf	keynum

	btfsc	howbyte, 006H	; flag prov. 10		
	bsf	keynum, 000H

	bcf	carrybit
	rrf	index, W
	addlw	003H
	movwf	adr_h
	
	movlw	002H
	subwf	keynum, W
	movwf	adr_l
	bcf	carrybit
	swapf 	adr_l, F	; moltiplica		
	rrf	adr_l, F	; per 8
	btfsc	index, 000H
	bsf	adr_l, 007H	; aggiunge 128 se necessario

	return


;-------------------------------------------------------------------
					; carica la key dalla eprom esterna e fa lo xor con date
keep_pkey_0	movlw	007H
		movwf	adr_h
		swapf	index, w
		btfsc	howbyte, 5
		goto	keep_pkey_01	; salta se settato il bit 5 (agg. PMK)
		
		addlw   080H            ; aggiunge offset PMK 00
		btfsc	howbyte, 006H	; flag prov. 10
		addlw   008H            ; aggiunge offset PMK 10		
		
keep_pkey_01	movwf	adr_l

keep_pkey	MOVF    adr_h,W         ; setta la pagina alta della eprom esterna
		MOVWF   PCLATH
		MOVF    adr_l,W         ; indirizzo basso
		MOVWF   EEADR		; temp1          
		MOVLW   tmpkeybuf
		MOVWF   FSR
		;MOVLW   008H
		;MOVWF   bytecnt
		call	load_8_cnt
		
		CALL    read_pkey       ; legge la pk
		MOVF    tmpkeybuf,W	; xor dei primi due byte della pk con date_h e date _l
		XORWF   date_h,F
		MOVF    (tmpkeybuf+1),W
		XORWF   date_l,F
		RETURN

;-------------------------------------------------------------------
					; legge la pkey
read_pkey	CALL    read_ees
		MOVWF   INDIR
		incf	FSR, F
		INCF    EEADR,F
		DECFSZ  bytecnt,F		; skip if Zero
		GOTO    read_pkey
		RETURN

;read_pkey	MOVF    temp1,W           
;		CALL    read_ee
;		MOVWF   INDIR
;		incf	FSR, F
;		INCF    temp1,F
;		DECFSZ  bytecnt,F		; skip if Zero
;		GOTO    read_pkey
;		RETURN

;****************************************************************************
;* IMPOSTAZIONI PER DECODIFICA KEY GENERICA
;****************************************************************************
seskeycrypt     call    keep_pkey
seskeycrypt2	bcf	howbyte, 7		;settare questo bit in caso di signature
		movlw	080H

;****************************************************************************
;* ROUTINE DI CRYPT ( IMPOSTARE BIT 7 DI HOWBYTE A 1 NEL CASO DI SIGNATURE )
;****************************************************************************

newcrypt	movwf	bytecnt
		clrf	bf3
		clrf	bf4

newcrypt_c1	movlw	00AH		;se bf3 = 10 azzeralo
		xorwf	bf3, W
		skipnz
		clrf	bf3

		movlw	010H		;se bf4 = 16 azzeralo
		xorwf	bf4, W
		skipnz
		clrf	bf4

		movf	bf3, F		;se bf3=0
		skpnz			;
		CALL    rot_right10	;ruota key10 di un bit a destra
		
		movlw	tmpkeybuf	;key10
		addwf	bf3, W
		movwf	FSR
		movf	INDIR, W
		movwf	bf5		;valore di d1(bf3)

		clrf	PCLATH		;pagina Hush Table 1
		btfss	bf5, 000H
		bsf	PCLATH, 000H	;pagina Hush Table 2

		movf	bf4, W
		call	offset_temp	;if W il valore di Key(bf4)		
		xorwf	bf5, W		;valore di Key(temp(bf4)) xor d1(bf3) in w
		
		call	read_ee		;legge valore dalla Tabella ( n. byte in W )		
		movwf	bf5		;e lo mette in bf5
		
		movlw	00FH		;se bf4 = 15 azzeralo
		xorwf	bf4, W
		skipz
		incf	bf4, W		;altrimenti aggiungi 1

		call	offset_temp	;in W il valore di Key(bf4+1)
		xorwf	bf5, W		;fa xor con il valore del precedente
		movwf	INDIR		;e lo rimemorizza
		
		incf	bf3, F
		incf	bf4, F

		decfsz	bytecnt, F
		goto	newcrypt_c1
		return
		
;****************************************************************************
;* ROUTINE DI APPOGGIO PER NEWCRYPT
;****************************************************************************
offset_temp	movwf	bf6
		andlw	007h
		btfss	bf6, 3
		goto 	tmp_key_val		
		btfsc	howbyte, 7
		goto	tmp_key_val

		bcf	carrybit
		movwf	bf6
		rlf	bf6, W
		addwf	bf6, W
		andlw	007h

tmp_key_val	addwf	temp2, W	;offset key
		movwf	FSR
		movf	INDIR, W
		return               
;-------------------------------------------------------------------

cardcrypt
	CALL    pagee_02           
	MOVLW   0B0H
	MOVWF   EEADR			; ex. bf7
	MOVLW   (tmpkeybuf + 7)
	MOVWF   FSR
	;MOVLW   008H
	;MOVWF   bytecnt
	call	load_8_cnt

cardcrypt_001
	CALL    read_ees
	MOVWF   INDIR
	DECF    FSR,F
	INCF    EEADR,F			; ex. bf7
	DECFSZ  bytecnt,F		; skip if Zero
	GOTO    cardcrypt_001
		
cardcrypt_001b
	call	copyleft_8
	
	;MOVLW   008H
	;MOVWF   bytecnt
	call	load_8_cnt

cardcrypt_003	
	call	rot_right8
	CLRF    bf8_lun
	
cardcrypt_004
	MOVLW   tmpkeybuf            
	MOVWF   FSR
	MOVF    bf8_lun,W
	ADDWF   FSR,F
	MOVF    INDIR,W
	MOVWF   bf7
	MOVF    temp2,W
	ADDLW   006H
	MOVWF   FSR
	MOVF    INDIR,W
	XORWF   bf7,F
	DECF    bytecnt,W
	XORWF   bf7,W
	CLRF    PCLATH
	CALL    read_ee
	MOVWF   bf7
	MOVF    temp2,W
	MOVWF   FSR
	MOVF    INDIR,W
	MOVWF   bf6
	MOVF    temp2,W
	MOVWF   FSR
	MOVF    bf7,W
	MOVWF   INDIR
	MOVLW   007H
	ADDWF   FSR,F
	MOVF    INDIR,W
	MOVWF   temp1
	MOVLW   007H
	SUBWF   FSR,F
	MOVF    temp1,W
	XORWF   INDIR,F
	MOVF    temp2,W
	MOVWF   FSR
	MOVLW   006H
	ADDWF   FSR,F
	MOVF    INDIR,W
	INCF    FSR,F
	MOVWF   INDIR
	DECF    FSR,F
	MOVF    bf7,W
	MOVWF   INDIR
	DECF    FSR,F
	MOVF    INDIR,W
	INCF    FSR,F
	XORWF   INDIR,F
	MOVLW   004H
	MOVWF   chksum

cardcrypt_005
	DECF    FSR,F           
	DECF    FSR,F
	MOVF    INDIR,W
	INCF    FSR,F
	MOVWF   INDIR
	DECFSZ  chksum,F		; skip if Zero
	GOTO    cardcrypt_005
	DECF    FSR,F
	MOVF    bf6,W
	MOVWF   INDIR
	INCF    bf8_lun,F
	BTFSS   bf8_lun,03H	; skip if bit = 1
	GOTO    cardcrypt_004

	DECFSZ  bytecnt,F		; skip if Zero
	GOTO    cardcrypt_003
	RETURN

;-----------------------------------------------------------------
					; chid_h e chid-l = chid byte alto e basso
sel_prov	MOVF    index,W
                MOVWF   bf4             ;copio index in bf4
                
		clrf    index           ;provider 0
		
		MOVLW   006H            ;punta alla pagina 6 dell'eeprom ( prov. 6 e 7 )
		MOVWF   PCLATH		;

		movlw	07FH		;80 - 1 ( offset keys del prov. 7 )
		movwf	EEADR

sp_ciclo1	;movlw	008H		;sostituire con 8 se si vogliono utilizzare
		;movwf	bytecnt		;otto ChId per ogni provider
					;( 2 pk del prov. n.8 per ogni provider )
		call	load_8_cnt
					
					
sp_ciclo2	incf	EEADR, F
		call	read_ees
		incf	EEADR, F
		
		xorwf	chid_h, W	;*** controlla parte alta
		skipz
		goto	sp_next_chid
				
		call	read_ees		
		
		xorwf	chid_l, W	;*** controlla parte bassa
		skipz
		goto	sp_next_chid

		goto	aggiorna_F9	;return
		
sp_next_chid	decfsz	bytecnt, F
		goto 	sp_ciclo2

		incf	index, F
		movf	index, W
		xorlw	007H
		skipz
		goto  	sp_ciclo1

		MOVF    bf4,W           
                MOVWF   index           ;rimango sul prov. attuale
                RETURN                

;-------------------------------------------------------------------
bigdelay				; ritardo
		IF SMD
		 MOVLW   01CH		; wafer smd
		ELSE
		 MOVLW   012H           ; wafer dip
		ENDIF

		movwf	EEADR
deley_1		CALL	delay
		decfsz	EEADR, F
		goto	deley_1
		RETURN

;-------------------------------------------------------------------
					; punta alla tabella delle risposte
pagee_02	movlw   002H            
pagee_02_001	movwf   PCLATH
		swapf	index, w
		return

;-------------------------------------------------------------------
					; punta all'ultima pagina delle eeprom esterna
pagee_07	movlw   007H            
		goto    pagee_02_001

;-------------------------------------------------------------------

read_in_sbuf	movwf	bytecnt
		movlw	sbuffer
read_in_sbuf1	movwf	FSR

read_in_sbuf2	call    pagee_02
		
read_in_sbufc	call	read_ees
		movwf	INDIR
		incf	FSR, F
		incf	EEADR, F

		decfsz	bytecnt, F
		goto	read_in_sbufc
		return
		
talk		movlw   008H
		movwf   EEADR
		movlw   bf2_cla - 1	; carica indirizzo buffer input/output
		movwf	FSR
		
talk_000	movlw   03FH            ; trasmette il primo byte a 01H
		movwf   chksum		; trasmette 8 byte da bf2_cla		
		movlw   001H
		goto	talk_002

send_sbuffer	movf	bf8_lun, W
		skipnz
		goto	talk_chk
		
		movwf	EEADR		; w numero di byte da inviare		
		movlw	sbuffer
		movwf	FSR
send_sbuffer2	call	talk_001

talk_chk	movf    chksum, W
		goto	txcam
		

talk_001	movf	INDIR,W         ; legge gli altri 7 dal buffer di I/O
talk_002	call	txcam
		incf	FSR,F
		decfsz	EEADR,F         ; skip if Zero
		goto	talk_001
		return
		
load_8_cnt	movlw	008H
		movwf	bytecnt
		return


;-----------------------------------------------------------
		
		
		ORG	2100H
		
		;       00,01,02,03,04,05,06,07,08,09,0A,0B,0C,0D,0E,0F
		DATA	01,05,11,1A,98,98,98,98,27,2C,32,74,98,98,7A,98
		
		;       10,11,12,13,14,15,16,17,18,19,1A,1B,1C,1D,1E,1F
		DATA	98,78,98,98,98,98,98,98,98,98,98,98,98,98,98,98

		
		DATA	01,79,00,00,00,00,00,00		;020H
		DATA	01,7C,00,00,00,00,00,00		;028H
		DATA	05,90,00,00,00,02,00,00		;030H
		
		
		END
